import 'dart:collection';
import 'dart:html';
import 'dart:math';
import 'dart:async';

import 'package:angular/angular.dart';

const SNACK_HOLDER_CLASS = 'snack-mr-holder';
const SNACK_BODY_CLASS = 'snack-mr-body';
const SNACK_APPEAR_CLASS = 'snack-mr-body--appear';
const SNACK_HIDE_CLASS = 'snack-mr-body--hide';
const SNACK_BLOCK_CLASS = 'snack-mr-block';
const SNACK_APPEAR_TIME = 250;
const SNACK_HIDE_TIME = 150;

@Injectable()
class SnackService {
  SnackConfig config;

  Element _root;
  Element _holder;

  List<SnackListElement> _list = <SnackListElement>[];

  SnackService();
  SnackService.fromConfig({this.config});

  void init() {
    if (config == null) {
      throw Exception('$selfString : No root component');
    }

    _root = querySelector(config.root);
    if (_root == null) {
      throw Exception('$selfString : Root element not found');
    }

    _holder = DivElement()
      ..className =
          '$SNACK_HOLDER_CLASS x-${config.xPosition.toString().split('.').last} y-${config.yPosition.toString().split('.').last} ${config.customHolderClasses}';

    _root.children.add(_holder);
  }

  void emit(SnackEmit em) {
    if (_root == null) {
      init();
    }

    var element = snackConstructor(em);
    var snLE = SnackListElement(emit: em, element: element);

    if (em.isBreaking) {
      _list.insert(0, snLE);
    } else {
      _list.add(snLE);
    }

    if (_list.length == 1) {
      _snackAppear(snLE);
    }
  }

  void _snackAppear(SnackListElement element) {
    _holder.children.add(element.element);
    element.element.classes.add(SNACK_APPEAR_CLASS);
    Future.delayed(const Duration(milliseconds: SNACK_APPEAR_TIME), () {
      element.element.classes.remove(SNACK_APPEAR_CLASS);
    });

    var timer =
        element.emit.timeout > 0 ? element.emit.timeout : config.timeout;
    Future.delayed(Duration(milliseconds: timer), () {
      try {
        // _snackHide(element.element);
      } catch (_) {}
    });
  }

  Future _snackHide(Element element) async {
    element.classes.add(SNACK_HIDE_CLASS);
    await Future.delayed(const Duration(milliseconds: SNACK_HIDE_TIME));

    try {
      if (_list.isNotEmpty) {
        var q = _list.firstWhere((x) => x.element == element);
        if (q != null) {
          _list.remove(q);
        }
      }

      element.remove();
    } catch (_) {}

    if (_list.isNotEmpty) {
      _snackAppear(_list.first);
    }
  }

  String get selfString => this.runtimeType.toString();
  Element snackConstructor(SnackEmit se) {
    var snackBlock = se.isBlock ? SNACK_BLOCK_CLASS : '';

    var body = DivElement()
      ..className = '$SNACK_BODY_CLASS ${config.customHolderClasses} $snackBlock'
      ..id = _randIndex;

    var textArea = DivElement()
      ..className = 'snack-mr-text'
      ..text = se.message;
    body.children.add(textArea);

    if (se.isAction || se.isDismiss) {
      var buttonHolder = DivElement()..className = 'snack-mr-buttons';

      if (se.isAction && se.actionCallback != null) {
        var action = ButtonElement()
          ..className = 'snack-mr-button-action'
          ..text = se.actionText;

        action.onClick.listen((e) => se.actionCallback());

        buttonHolder.children.add(action);
      }

      if (se.isDismiss) {
        var dismiss = ButtonElement()
          ..className = 'snack-mr-button-dismiss'
          ..text = se.dismissText;

        if (se.dismissCallback != null) {
          var dCallback = se.dismissCallback;
          se.dismissCallback = () {
            _snackHide(body);
            dCallback();
          };
        } else {
          se.dismissCallback = () => _snackHide(body);
        }

        dismiss.onClick.listen((e) => se.dismissCallback());
        buttonHolder.children.add(dismiss);
      }

      body.children.add(buttonHolder);
    }

    return body;
  }

  Random _rand;
  String get _randIndex {
    if (_rand == null) {
      _rand = Random();
    }
    var codeUnits = List.generate(8, (index) {
      return _rand.nextInt(33) + 89;
    });

    return String.fromCharCodes(codeUnits);
  }
}

class SnackConfig {
  String root = 'body';
  int timeout = 15000;
  SnackXPositionMode xPosition = SnackXPositionMode.RIHT;
  SnackYPositionMode yPosition = SnackYPositionMode.BOTTOM;
  String customHolderClasses = '';
  String customBodyClasses = '';
  String customButtonClasses = '';

  SnackConfig(
      {this.root,
      this.timeout,
      this.xPosition,
      this.yPosition,
      this.customHolderClasses,
      this.customBodyClasses,
      this.customButtonClasses});

  SnackConfig.empty();
}

class SnackListElement {
  Element element;
  SnackEmit emit;

  SnackListElement({this.element, this.emit});
}

class SnackEmit {
  String message;
  int timeout = -1;
  bool isBreaking = false;

  bool isAction = false;
  String actionText = '';
  Function actionCallback = null;

  bool isDismiss = true;
  String dismissText = 'Dismiss';
  Function dismissCallback = null;

  bool isBlock = false;

  SnackEmit(
      {this.message,
      this.timeout,
      this.isBreaking,
      this.actionText,
      this.isAction,
      this.actionCallback,
      this.isDismiss,
      this.dismissCallback,
      this.dismissText,
      this.isBlock});

  SnackEmit.simple(String message) {
    this.message = message;
  }

  SnackEmit.simpleBlock(String message){
    this.message = message;
    this.isBlock = true;
  }

  SnackEmit.action(String message, Function callback, String actionText,
      [bool isDissmiss = false]) {
    timeout = -1;
    this.isDismiss = isDissmiss;
    this.message = message;
    this.actionCallback = callback;
    isAction = true;
    this.actionText = actionText;
  }

  SnackEmit.actionBlock(String message, Function callback, String actionText,
      [bool isDissmiss = false]) {
    timeout = -1;
    this.isDismiss = isDissmiss;
    this.message = message;
    this.actionCallback = callback;
    isAction = true;
    this.actionText = actionText;
    this.isBlock = true;
  }
}

enum SnackXPositionMode { LEFT, CENTER, RIHT }

enum SnackYPositionMode { BOTTOM, TOP }
